/****************************************************************************
**
** Copyright (C) 2015 The Qt Company Ltd.
** Contact: http://www.qt.io/licensing/
**
** This file is part of the Qt Messaging Framework.
**
** $QT_BEGIN_LICENSE:LGPL21$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see http://www.qt.io/terms-conditions. For further
** information use the contact form at http://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 or version 3 as published by the Free
** Software Foundation and appearing in the file LICENSE.LGPLv21 and
** LICENSE.LGPLv3 included in the packaging of this file. Please review the
** following information to ensure the GNU Lesser General Public License
** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** As a special exception, The Qt Company gives you certain additional
** rights. These rights are described in The Qt Company LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef STATUSMONITOR_H
#define STATUSMONITOR_H

#include <QObject>
#include <qmailserviceaction.h>

class StatusMonitor;
class QMailServiceAction;

class StatusItem : public QObject
{
    Q_OBJECT

signals:
    void finished();
    void progressChanged(uint,uint);
    void statusChanged(const QString& message);

public:
    virtual QWidget* widget() = 0;
    virtual QPair<uint,uint> progress() const = 0;
    virtual QString status() const = 0;
};

class ServiceActionStatusItem : public StatusItem
{
    Q_OBJECT

public:
    ServiceActionStatusItem(QMailServiceAction* sa, const QString& description);
    QWidget* widget();
    QPair<uint,uint> progress() const;
    QString status() const;

private slots:
    void serviceActivityChanged(QMailServiceAction::Activity a);
    void serviceStatusChanged(const QMailServiceAction::Status& s);

private:
    QMailServiceAction* m_serviceAction;
    QString m_description;
};

class StatusMonitor : public QObject
{
    Q_OBJECT

public:
    StatusMonitor();
    static StatusMonitor* instance();
    void add(StatusItem* newItem);
    int itemCount() const;

signals:
    void added(StatusItem* s);
    void removed(StatusItem* s);
    void progressChanged(uint,uint);
    void statusChanged(const QString& msg);

private slots:
    void statusItemFinished();
    void statusItemProgressChanged();
    void statusItemDestroyed(QObject*);

private:
    void updateProgress();

private:
    QList<StatusItem*> m_statusItems;
};

#endif
