/****************************************************************************
**
** Copyright (C) 2015 The Qt Company Ltd.
** Contact: http://www.qt.io/licensing/
**
** This file is part of the Qt Messaging Framework.
**
** $QT_BEGIN_LICENSE:LGPL21$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see http://www.qt.io/terms-conditions. For further
** information use the contact form at http://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 or version 3 as published by the Free
** Software Foundation and appearing in the file LICENSE.LGPLv21 and
** LICENSE.LGPLv3 included in the packaging of this file. Please review the
** following information to ensure the GNU Lesser General Public License
** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** As a special exception, The Qt Company gives you certain additional
** rights. These rights are described in The Qt Company LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "listfilterplugin.h"
#include <qmaildisconnected.h>
#include <QtPlugin>

Q_EXPORT_PLUGIN2(listfilterplugin,ListFilterPlugin)

QMailStore::ErrorCode ListFilterPlugin::ListContentManager::add(QMailMessage *message, DurabilityRequirement)
{
    // is this message sent directly?
    QMailAccountKey accountKey;
    foreach(QMailAddress recip , message->recipients())
        accountKey |= QMailAccountKey::fromAddress(recip.address());

    if ( QMailStore::instance()->countAccounts(accountKey) )
        return QMailStore::NoError; // Don't filter messages sent directly

    QString listId(message->listId());
    if (!listId.isEmpty()) {
        QMailFolderIdList folderIds(QMailStore::instance()->queryFolders(QMailFolderKey::displayName(listId)
                                                                              & QMailFolderKey::parentAccountId(message->parentAccountId())));
        if (!folderIds.isEmpty()) {
            QMailFolderId folderId(folderIds.first());
            if (folderId != message->parentFolderId()) {
                QMailDisconnected::moveToFolder(message, folderId);
            }
        }
    }

    return QMailStore::NoError;
}

QString ListFilterPlugin::key() const
{
    return QString("listfilterplugin");
}

ListFilterPlugin::ListContentManager *ListFilterPlugin::create()
{
    return &manager;
}
