/****************************************************************************
**
** Copyright (C) 2018 Caliste Damien.
** Contact: Damien Caliste <dcaliste@free.fr>
**
** Copyright (C) 2018 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Messaging Framework.
**
** $QT_BEGIN_LICENSE:LGPL21$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see http://www.qt.io/terms-conditions. For further
** information use the contact form at http://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 or version 3 as published by the Free
** Software Foundation and appearing in the file LICENSE.LGPLv21 and
** LICENSE.LGPLv3 included in the packaging of this file. Please review the
** following information to ensure the GNU Lesser General Public License
** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** As a special exception, The Qt Company gives you certain additional
** rights. These rights are described in The Qt Company LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef QMAILCRYPTOFWD_H
#define QMAILCRYPTOFWD_H

#include "qmailglobal.h"
#include <QMap>
#include <QVariant>

class QMF_EXPORT QMailCryptoFwd
{
public:
    enum SignatureResult {
        SignatureValid,
        SignatureExpired,
        KeyExpired,
        CertificateRevoked,
        BadPassphrase,
        BadSignature,
        MissingKey,
        MissingSignature,
        UnusableKey,
        UnknownError
    };
    enum CryptResult {
        NoDigitalEncryption,
        Decrypted,
        WrongPassphrase,
        UnsupportedProtocol,
        UnknownCryptError
    };
    typedef QString (*PassphraseCallback)(const QString &info);

    struct QMF_EXPORT KeyResult {
        QString key;
        SignatureResult status;
        QVariantMap details;

        KeyResult(const QString &fingerprint, SignatureResult result,
                  const QVariantMap &parameters = QVariantMap())
            : key(fingerprint), status(result), details(parameters)
        {
        }
    };

    struct QMF_EXPORT VerificationResult {
        QString engine;
        SignatureResult summary;
        QList<KeyResult> keyResults;

        VerificationResult(SignatureResult status = UnknownError,
                           const QList<KeyResult> &results = QList<KeyResult>())
            : summary(status), keyResults(results)
        {
        }
    };

    struct QMF_EXPORT DecryptionResult {
        QString engine;
        CryptResult status;

        DecryptionResult(CryptResult result = UnknownCryptError)
            : status(result)
        {
        }
    };
};

#endif
