/****************************************************************************
**
** Copyright (C) 2015 The Qt Company Ltd.
** Contact: http://www.qt.io/licensing/
**
** This file is part of the Qt Messaging Framework.
**
** $QT_BEGIN_LICENSE:LGPL21$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see http://www.qt.io/terms-conditions. For further
** information use the contact form at http://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 or version 3 as published by the Free
** Software Foundation and appearing in the file LICENSE.LGPLv21 and
** LICENSE.LGPLv3 included in the packaging of this file. Please review the
** following information to ensure the GNU Lesser General Public License
** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** As a special exception, The Qt Company gives you certain additional
** rights. These rights are described in The Qt Company LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef SEARCHVIEW_H
#define SEARCHVIEW_H

#include <QMainWindow>
#include <qmailmessagekey.h>
#include <qmailserviceaction.h>

class MessageListView;
class SearchTermsComposer;
class FolderSelectorWidget;
class SearchButton;
class BodySearchWidget;
class QMailSearchAction;

QT_BEGIN_NAMESPACE

class QCheckBox;
class QLineEdit;
class QPushButton;
class QRadioButton;
class QStatusBar;
class QToolButton;

QT_END_NAMESPACE

class SearchView : public QMainWindow
{
    Q_OBJECT

public:
    SearchView(QWidget * parent = Q_NULLPTR, Qt::WindowFlags flags = Qt::Widget);
    void setVisible(bool visible);

signals:
    void searchResultSelected(const QMailMessageId& id);

public slots:
    void reset();
    void close();

private:
    void setupUi();
    QMailMessageKey searchKey() const;

private slots:
    void startSearch();
    void stopSearch();
    void messageIdsMatched(const QMailMessageIdList& ids);
    void searchProgressChanged(uint value, uint total);
    void searchActivityChanged(QMailServiceAction::Activity a);

private:
    void setupSearchAction();

    SearchButton* m_searchButton;
    QPushButton* m_resetButton;
    QPushButton* m_closeButton;
    FolderSelectorWidget* m_folderSelectorWidget;
    SearchTermsComposer* m_searchTermsComposer;
    BodySearchWidget* m_bodySearchWidget;
    MessageListView* m_searchResults;
    QMailSearchAction* m_searchAction;
    QStatusBar* m_statusBar;
    bool m_searched;
};

#endif
